import platform
from pathlib import Path
from unittest import TestCase, skipUnless
from unittest.mock import call, patch

from dakara_player.font import (
    FontLoaderLinux,
    FontLoaderNotAvailableError,
    FontLoaderWindows,
    get_font_loader_class,
)


class GetFontLoaderClassTestCase(TestCase):
    """Test font loader class getter."""

    def test(self):
        """Test to get the correct font loader class for the platform."""
        # call for Linux
        with patch("dakara_player.font.platform.system", return_value="Linux"):
            FontLoaderClass = get_font_loader_class()
            self.assertEqual(FontLoaderClass, FontLoaderLinux)

        # call for Windows
        with patch("dakara_player.font.platform.system", return_value="Windows"):
            FontLoaderClass = get_font_loader_class()
            self.assertEqual(FontLoaderClass, FontLoaderWindows)

        # call for uniplemented OS
        with patch("dakara_player.font.platform.system", return_value="other"):
            with self.assertRaisesRegex(
                NotImplementedError,
                r"This operating system \(other\) is not currently supported",
            ):
                get_font_loader_class()


class FontLoaderTestCase(TestCase):
    """Helper to test font loader classes."""

    def setUp(self):
        # create directory
        self.directory = Path("directory")

        # create font file list
        self.font_name = "font_file.ttf"
        self.font_name_list = [self.font_name]
        self.font_path = self.directory / self.font_name
        self.font_path_list = {self.font_name: self.font_path}


class FontLoaderCommonTestCase(FontLoaderTestCase):
    """Test common methods of the font loaders."""

    def get_font_loader(self):
        """Return an instance of the font loader."""
        with self.assertLogs("dakara_player.font", "DEBUG"):
            return FontLoaderLinux("package")

    @patch.object(FontLoaderLinux, "unload", autospec=True)
    def test_context_manager(self, mocked_unload):
        """Test the font loader context manager."""
        with FontLoaderLinux("package") as loader:
            pass

        mocked_unload.assert_called_with(loader)

    @patch.object(Path, "is_file", autospec=True)
    @patch("dakara_player.font.files", autospec=True)
    def test_get_font_name_list(self, mocked_files, mocked_is_file):
        """Test to get list of font names."""
        # mock system calls
        mocked_files.return_value.iterdir.return_value = [
            Path("dir") / self.font_name,
            Path("dir") / "__init__.py",
        ]
        mocked_is_file.return_value = True

        font_loader = self.get_font_loader()

        # call the method
        with self.assertLogs("dakara_player.font", "DEBUG") as logger:
            font_name_list = font_loader.get_font_name_list()

        # assert the result
        self.assertListEqual(font_name_list, self.font_name_list)

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.font:Scanning fonts directory",
                "DEBUG:dakara_player.font:Found 1 font(s) to load",
            ],
        )

        # assert mock
        mocked_files.assert_called_once_with("package")

    @patch("dakara_player.font.as_file", autospec=True)
    @patch("dakara_player.font.files", autospec=True)
    @patch.object(FontLoaderLinux, "get_font_name_list", autospec=True)
    def test_get_font_path_iterator(
        self, mocked_get_font_name_list, mocked_files, mocked_as_file
    ):
        """Test to get iterator of font paths."""
        mocked_get_font_name_list.return_value = self.font_name_list
        mocked_as_file.return_value.__enter__.return_value = (
            self.directory / self.font_name
        )

        font_loader = self.get_font_loader()

        # call the method
        font_file_path_list = list(font_loader.get_font_path_iterator())

        # assert the result
        self.assertListEqual(font_file_path_list, [self.font_path])

        # assert mock
        mocked_files.assert_called_once_with("package")
        mocked_files.return_value.joinpath.assert_called_once_with(self.font_name)


@skipUnless(platform.system() == "Linux", "Can be tested on Linux only")
class FontLoaderLinuxTestCase(FontLoaderTestCase):
    """Test the Linux font loader."""

    def setUp(self):
        super().setUp()

        # save user directory
        self.user_directory = Path("~").expanduser()

    def get_font_loader(self):
        """Return an instance of the font loader."""
        with self.assertLogs("dakara_player.font", "DEBUG"):
            return FontLoaderLinux("package")

    @patch.object(FontLoaderLinux, "load_font", autospec=True)
    @patch.object(FontLoaderLinux, "get_font_path_iterator", autospec=True)
    @patch.object(Path, "rglob", autospec=True)
    @patch.object(Path, "mkdir", autospec=True)
    def test_load(
        self,
        mocked_mkdir,
        mocked_rglob,
        mocked_get_font_path_iterator,
        mocked_load_font,
    ):
        """Test to load fonts."""
        # prepare the mock
        mocked_get_font_path_iterator.return_value = (p for p in [self.font_path])
        mocked_rglob.side_effect = [
            [Path("/") / "usr" / "share" / "fonts" / "font1"],
            [self.user_directory / ".fonts" / "font2"],
        ]

        font_loader = self.get_font_loader()

        # call the method
        font_loader.load()

        # assert the call
        mocked_mkdir.assert_called_once_with(
            self.user_directory / ".fonts", parents=True, exist_ok=True
        )
        mocked_rglob.assert_has_calls(
            [
                call(Path("/") / "usr" / "share" / "fonts", "*"),
                call(self.user_directory / ".fonts", "*"),
            ]
        )
        mocked_get_font_path_iterator.assert_called_once_with(font_loader)
        mocked_load_font.assert_called_once_with(
            font_loader,
            self.font_path,
            ["font1"],
            ["font2"],
        )

    @patch.object(Path, "unlink", autospec=True)
    @patch("dakara_player.font.copy", autospec=True)
    @patch.object(Path, "is_symlink", autospec=True)
    def test_load_font_system(self, mocked_is_symlink, mocked_copy, mocked_unlink):
        """Test to load one font which is in system directory."""
        font_loader = self.get_font_loader()

        # pre assertions
        self.assertEqual(len(font_loader.fonts_loaded), 0)

        # call the method
        with self.assertLogs("dakara_player.font", "DEBUG") as logger:
            font_loader.load_font(
                self.font_path,
                ["font_file.ttf"],
                [],
            )

        # post assertions
        self.assertEqual(len(font_loader.fonts_loaded), 0)

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.font:Font 'font_file.ttf' "
                "found in system directory"
            ],
        )

        # assert the call
        mocked_is_symlink.assert_not_called()
        mocked_unlink.assert_not_called()
        mocked_copy.assert_not_called()

    @patch.object(Path, "unlink", autospec=True)
    @patch("dakara_player.font.copy", autospec=True)
    @patch.object(Path, "is_symlink", autospec=True)
    def test_load_font_user(self, mocked_is_symlink, mocked_copy, mocked_unlink):
        """Test to load one font which is in user directory."""
        font_loader = self.get_font_loader()

        # pre assertions
        self.assertEqual(len(font_loader.fonts_loaded), 0)

        # call the method
        with self.assertLogs("dakara_player.font", "DEBUG") as logger:
            font_loader.load_font(
                self.font_path,
                [],
                ["font_file.ttf"],
            )

        # post assertions
        self.assertEqual(len(font_loader.fonts_loaded), 0)

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.font:Font 'font_file.ttf' "
                "found in user directory"
            ],
        )

        # assert the call
        mocked_is_symlink.assert_not_called()
        mocked_unlink.assert_not_called()
        mocked_copy.assert_not_called()

    @patch.object(Path, "unlink", autospec=True)
    @patch("dakara_player.font.copy", autospec=True)
    @patch.object(Path, "is_symlink", autospec=True)
    def test_load_font_user_link_dead_install(
        self,
        mocked_is_symlink,
        mocked_copy,
        mocked_unlink,
    ):
        """Test to load one font which is in user directory as dead link."""
        # prepare the mock
        mocked_is_symlink.return_value = True

        font_loader = self.get_font_loader()

        # pre assertions
        self.assertEqual(len(font_loader.fonts_loaded), 0)

        # call the method
        with self.assertLogs("dakara_player.font", "DEBUG") as logger:
            font_loader.load_font(self.font_path, [], [])

        # post assertions
        self.assertEqual(len(font_loader.fonts_loaded), 1)

        # assert effect on logs
        font_path = self.user_directory / ".fonts/font_file.ttf"
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.font:Dead symbolic link found for "
                "font 'font_file.ttf' in user directory, removing it",
                "DEBUG:dakara_player.font:Font 'font_file.ttf' "
                "loaded in user directory: '{}'".format(font_path),
            ],
        )

        # assert the call
        mocked_is_symlink.assert_called_with(font_path)
        mocked_unlink.assert_called_with(font_path, missing_ok=True)
        mocked_copy.assert_called_once_with(
            Path("directory") / "font_file.ttf", font_path
        )

    @patch.object(Path, "unlink", autospec=True)
    @patch("dakara_player.font.copy", autospec=True)
    @patch.object(Path, "is_symlink", autospec=True)
    def test_load_font_install(self, mocked_is_symlink, mocked_copy, mocked_unlink):
        """Test to load one font which is not installed."""
        # prepare the mock
        mocked_is_symlink.return_value = False

        font_loader = self.get_font_loader()

        # pre assertions
        self.assertEqual(len(font_loader.fonts_loaded), 0)

        # call the method
        with self.assertLogs("dakara_player.font", "DEBUG") as logger:
            font_loader.load_font(self.font_path, [], [])

        # post assertions
        self.assertEqual(len(font_loader.fonts_loaded), 1)
        self.assertEqual(
            font_loader.fonts_loaded[self.font_name],
            self.user_directory / ".fonts/font_file.ttf",
        )

        # assert effect on logs
        font_path = self.user_directory / ".fonts/font_file.ttf"
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.font:Font 'font_file.ttf' "
                "loaded in user directory: '{}'".format(font_path)
            ],
        )

        # assert the call
        mocked_is_symlink.assert_called_with(font_path)
        mocked_unlink.assert_not_called()
        mocked_copy.assert_called_once_with(
            Path("directory") / "font_file.ttf",
            self.user_directory / ".fonts/font_file.ttf",
        )

    @patch.object(Path, "unlink", autospec=True)
    def test_unload(self, mocked_unlink):
        """Test to unload fonts."""
        font_loader = self.get_font_loader()

        # set a font as loaded
        font_loader.fonts_loaded = {"font1": Path("font1"), "font2": Path("font2")}

        # call the method
        font_loader.unload()

        # assert the call
        # especially check that the unload function does not alter the list of
        # elements we are iterating on
        mocked_unlink.assert_has_calls([call(Path("font1")), call(Path("font2"))])

    @patch.object(Path, "unlink", autospec=True)
    def test_unload_font(self, mocked_unlink):
        """Test to unload one font."""
        font_loader = self.get_font_loader()

        # set a font as loaded
        font_loader.fonts_loaded = self.font_path_list

        # pre assert there is one font loaded
        self.assertEqual(len(font_loader.fonts_loaded), 1)

        # call the method
        with self.assertLogs("dakara_player.font", "DEBUG") as logger:
            font_loader.unload_font(self.font_name)

        # assert there are no font loaded anymore
        self.assertEqual(len(font_loader.fonts_loaded), 0)

        # assert effect of logs
        self.assertListEqual(
            logger.output,
            ["DEBUG:dakara_player.font:Font 'font_file.ttf' unloaded"],
        )

        # assert the call
        mocked_unlink.assert_called_once_with(self.font_path)

    @patch.object(Path, "unlink", autospec=True)
    def test_unload_font_error(self, mocked_unlink):
        """Test to unload one font when unable to remove font."""
        mocked_unlink.side_effect = OSError("error message")

        font_loader = self.get_font_loader()

        # set a font as loaded
        font_loader.fonts_loaded = self.font_path_list

        # pre assert there is one font loaded
        self.assertEqual(len(font_loader.fonts_loaded), 1)

        # call the method
        with self.assertLogs("dakara_player.font", "ERROR") as logger:
            font_loader.unload_font(self.font_name)

        # assert there are no font loaded anymore
        self.assertEqual(len(font_loader.fonts_loaded), 0)

        # assert effect of logs
        self.assertListEqual(
            logger.output,
            [
                "ERROR:dakara_player.font:Font 'font_file.ttf' in "
                "'directory/font_file.ttf' cannot be unloaded: error message"
            ],
        )


class FontLoaderWindowsTestCase(TestCase):
    """Test the Windows font loader."""

    def setUp(self):
        # create directory
        self.directory = Path("directory")

        # create font file list
        self.font_name = "font_file.ttf"
        self.font_name_list = [self.font_name]
        self.font_path = self.directory / self.font_name
        self.font_path_list = {self.font_name: self.font_path}

    def get_font_loader(self):
        """Return an instance of the font loader."""
        with self.assertLogs("dakara_player.font", "DEBUG"):
            return FontLoaderWindows("package")

    @patch("dakara_player.font.ctypes", spec=[])
    def test_init_no_windll(self, mocked_ctypes):
        """Test to create font loader on a different OS."""
        with self.assertRaisesRegex(
            FontLoaderNotAvailableError, "FontLoaderWindows can only be used on Windows"
        ):
            self.get_font_loader()

    @patch("dakara_player.font.ctypes")
    @patch.object(FontLoaderWindows, "load_font", autospec=True)
    @patch.object(FontLoaderWindows, "get_font_path_iterator", autospec=True)
    def test_load(self, mocked_get_font_path_iterator, mocked_load_font, mocked_ctypes):
        """Test to load fonts."""
        # prepare the mock
        mocked_get_font_path_iterator.return_value = (p for p in [self.font_path])

        font_loader = self.get_font_loader()

        # call the method
        font_loader.load()

        # assert the call
        mocked_get_font_path_iterator.assert_called_once_with(font_loader)
        mocked_load_font.assert_called_once_with(font_loader, self.font_path)

    @patch("dakara_player.font.ctypes")
    def test_load_font(self, mocked_ctypes):
        """Test to load one font."""
        # setup mock
        mocked_ctypes.windll.gdi32.AddFontResourceW.return_value = 1

        font_loader = self.get_font_loader()

        # pre assertions
        self.assertEqual(len(font_loader.fonts_loaded), 0)

        # call the method
        with self.assertLogs("dakara_player.font", "DEBUG") as logger:
            font_loader.load_font(self.font_path)

        # post assertions
        self.assertEqual(len(font_loader.fonts_loaded), 1)

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            ["DEBUG:dakara_player.font:Font 'font_file.ttf' loaded"],
        )

        # assert the call
        mocked_ctypes.windll.gdi32.AddFontResourceW.assert_called_once_with(
            self.font_path
        )

    @patch("dakara_player.font.ctypes")
    def test_load_font_error(self, mocked_ctypes):
        """Test to fail to load one font."""
        # setup mock
        mocked_ctypes.windll.gdi32.AddFontResourceW.return_value = 0

        font_loader = self.get_font_loader()

        # pre assertions
        self.assertEqual(len(font_loader.fonts_loaded), 0)

        # call the method
        with self.assertLogs("dakara_player.font", "DEBUG") as logger:
            font_loader.load_font(self.font_path)

        # post assertions
        self.assertEqual(len(font_loader.fonts_loaded), 0)

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            ["WARNING:dakara_player.font:Font 'font_file.ttf' cannot be loaded"],
        )

    @patch("dakara_player.font.ctypes")
    @patch.object(FontLoaderWindows, "unload_font", autospec=True)
    def test_unload(self, mocked_unload_font, mocked_ctypes):
        """Test to unload fonts."""
        font_loader = self.get_font_loader()

        # set a font as loaded
        font_loader.fonts_loaded = {"font1": Path("font1"), "font2": Path("font2")}

        # call the method
        font_loader.unload()

        # assert the call
        mocked_unload_font.assert_has_calls(
            [call(font_loader, "font1"), call(font_loader, "font2")]
        )

    @patch("dakara_player.font.ctypes")
    def test_unload_font(self, mocked_ctypes):
        """Test to unload one font."""
        # setup mock
        mocked_ctypes.windll.gdi32.RemoveFontResourceW.return_value = 1

        font_loader = self.get_font_loader()

        # set a font as loaded
        font_loader.fonts_loaded = self.font_path_list

        # pre assert there is one font loaded
        self.assertEqual(len(font_loader.fonts_loaded), 1)

        # call the method
        with self.assertLogs("dakara_player.font", "DEBUG") as logger:
            font_loader.unload_font(self.font_name)

        # assert there are no font loaded anymore
        self.assertEqual(len(font_loader.fonts_loaded), 0)

        # assert effect of logs
        self.assertListEqual(
            logger.output,
            ["DEBUG:dakara_player.font:Font 'font_file.ttf' unloaded"],
        )

        # assert the call
        mocked_ctypes.windll.gdi32.RemoveFontResourceW.assert_called_once_with(
            self.font_path
        )

    @patch("dakara_player.font.ctypes")
    def test_unload_font_error(self, mocked_ctypes):
        """Test to fail to unload one font."""
        # setup mock
        mocked_ctypes.windll.gdi32.RemoveFontResourceW.return_value = 0

        font_loader = self.get_font_loader()

        # set a font as loaded
        font_loader.fonts_loaded = self.font_path_list

        # pre assert there is one font loaded
        self.assertEqual(len(font_loader.fonts_loaded), 1)

        # call the method
        with self.assertLogs("dakara_player.font", "DEBUG") as logger:
            font_loader.unload_font(self.font_name)

        # assert there are no font loaded anymore
        self.assertEqual(len(font_loader.fonts_loaded), 0)

        # assert effect of logs
        self.assertListEqual(
            logger.output,
            [
                "WARNING:dakara_player.font:Font 'font_file.ttf' cannot be unloaded",
            ],
        )



"""Manage the player."""

import logging

logger = logging.getLogger("dakara_manager")


class DakaraManager:
    """Manager for the Dakara player.

    This object is a high-level manager for the Dakara player. It controls the
    different elements of the project with simple commands.

    Args:
        font_loader (font.FontLoader): Object for font
            installation/deinstallation.
        media_player (media_player.MediaPlayer): Interface to the media player.
        client_http (web_client.HTTPClientDakara): Client for the Dakara server
            for the HTTP protocol.
        client_websocket (web_client.WebSocketClientDakara): Client for the
            Dakara server for the Websocket protocol.
    """

    def __init__(self, font_loader, media_player, client_http, client_websocket):
        # set modules up
        self.font_loader = font_loader
        self.media_player = media_player
        self.client_http = client_http
        self.client_websocket = client_websocket

        # set player callbacks
        self.media_player.set_callback(
            "started_transition", self.handle_started_transition
        )
        self.media_player.set_callback("started_song", self.handle_started_song)
        self.media_player.set_callback("could_not_play", self.handle_could_not_play)
        self.media_player.set_callback("finished", self.handle_finished)
        self.media_player.set_callback("paused", self.handle_paused)
        self.media_player.set_callback("resumed", self.handle_resumed)
        self.media_player.set_callback("error", self.handle_error)
        self.media_player.set_callback("updated_timing", self.handle_updated_timing)

        # set dakara server websocket callbacks
        self.client_websocket.set_callback("idle", self.play_idle_screen)
        self.client_websocket.set_callback("playlist_entry", self.play_playlist_entry)
        self.client_websocket.set_callback("command", self.do_command)
        self.client_websocket.set_callback("connection_lost", self.play_idle_screen)

    def handle_error(self, playlist_entry_id, message):
        """Callback when a media player error occurs.

        Args:
            playlist_entry_id (int): Playlist entry ID.
            message (str): Text describing the error.
        """
        self.client_http.post_player_error(playlist_entry_id, message)

    def handle_finished(self, playlist_entry_id):
        """Callback when a playlist entry finishes.

        Args:
            playlist_entry_id (int): Playlist entry ID.
        """
        self.client_http.put_status_finished(playlist_entry_id)

    def handle_started_transition(self, playlist_entry_id):
        """Callback when the transition of a playlist entry starts.

        Args:
            playlist_entry_id (int): Playlist entry ID.
        """
        self.client_http.put_status_started_transition(playlist_entry_id)

    def handle_started_song(self, playlist_entry_id):
        """Callback when the song of a playlist entry starts.

        Args:
            playlist_entry_id (int): Playlist entry ID.
        """
        self.client_http.put_status_started_song(playlist_entry_id)

    def handle_could_not_play(self, playlist_entry_id):
        """Callback when a playlist entry could not play.

        Args:
            playlist_entry_id (int): Playlist entry ID.
        """
        self.client_http.put_status_could_not_play(playlist_entry_id)

    def handle_paused(self, playlist_entry_id, timing):
        """Callback when the player is paused.

        Args:
            playlist_entry_id (int): Playlist entry ID.
            timing (int): Position of the player in seconds.
        """
        self.client_http.put_status_paused(playlist_entry_id, timing)

    def handle_resumed(self, playlist_entry_id, timing):
        """Callback when the player resumed playing.

        Args:
            playlist_entry_id (int): Playlist entry ID.
            timing (int): Position of the player in seconds.
        """
        self.client_http.put_status_resumed(playlist_entry_id, timing)

    def handle_updated_timing(self, playlist_entry_id, timing):
        """Callback when the player updated its timing.

        Args:
            playlist_entry_id (int): Playlist entry ID.
            timing (int): Position of the player in seconds.
        """
        self.client_http.put_status_update_timing(playlist_entry_id, timing)

    def play_playlist_entry(self, playlist_entry):
        """Play the requested playlist entry.

        Args:
            playlist_entry (dict): Dictionary of the playlist entry.
        """
        self.media_player.set_playlist_entry(playlist_entry)

    def play_idle_screen(self):
        """Play the idle screen."""
        # skip currently playing file if any
        self.media_player.skip(no_callback=True)

        self.media_player.play("idle")

    def do_command(self, command):
        """Execute a player command.

        Args:
            command (str): Name of the command to execute.

        Raises:
            AssertError: If the command is not known.
        """
        assert command in (
            "pause",
            "resume",
            "restart",
            "skip",
            "rewind",
            "fast_forward",
        ), "Unknown command requested: '{}'".format(command)

        getattr(self.media_player, command)()



from importlib.resources import as_file, files
from pathlib import Path
from pathlib import Path as Path_pathlib
from shutil import copy
from tempfile import TemporaryDirectory
from unittest import TestCase
from unittest.mock import MagicMock, patch

from dakara_player.text import (
    TemplateNotFoundError,
    TextGenerator,
    separate_package_last_directory,
)


class TextGeneratorTestCase(TestCase):
    """Test the text generator class unitary."""

    @patch.object(TextGenerator, "load_templates")
    @patch.object(TextGenerator, "load_icon_map")
    def test_load(self, mocked_load_icon_map, mocked_load_templates):
        """Test the load method."""
        # create ojbect
        text_generator = TextGenerator("package")

        # call the method
        text_generator.load()

        # assert the call
        mocked_load_icon_map.assert_called_once_with()
        mocked_load_templates.assert_called_once_with()

    @patch.object(Path_pathlib, "read_text", autospec=True)
    @patch("dakara_player.text.json.loads", autospec=True)
    def test_load_icon_map(self, mocked_loads, mocked_read_text):
        """Test to load the icon map."""
        # create the mock
        mocked_loads.return_value = {"name": "value"}
        mocked_read_text.return_value = '{"name": "value"}'

        # create the object
        text_generator = TextGenerator("package")

        # pre assert there are not icon map
        self.assertDictEqual(text_generator.icon_map, {})

        # call the method
        text_generator.load_icon_map()

        # assert there is an icon map
        self.assertDictEqual(text_generator.icon_map, {"name": "value"})

        # assert the mock
        mocked_loads.assert_called_with(mocked_read_text.return_value)

    @patch.object(TextGenerator, "check_template", autospec=True)
    @patch("dakara_player.text.ChoiceLoader", autospec=True)
    @patch("dakara_player.text.PackageLoader", autospec=True)
    @patch("dakara_player.text.FileSystemLoader", autospec=True)
    @patch("dakara_player.text.Environment", autospec=True)
    def test_load_templates(
        self,
        mocked_environment_class,
        mocked_file_system_loader_class,
        mocked_package_loader_class,
        mocked_choice_loader_class,
        mocked_check_template,
    ):
        """Test to load templates for text."""
        mocked_environment_class.return_value.filters = {}
        # create object
        text_generator = TextGenerator(
            "package.templates", Path("directory"), filenames={"text": "text.ass"}
        )

        # pre assert there are no templates
        self.assertIsNone(text_generator.environment)

        # call the method
        text_generator.load_templates()

        # assert there are templates defined
        self.assertIsNotNone(text_generator.environment)

        mocked_file_system_loader_class.assert_called_with(Path("directory"))
        mocked_package_loader_class.assert_called_with("package", "templates")
        mocked_choice_loader_class.assert_called_with(
            [
                mocked_file_system_loader_class.return_value,
                mocked_package_loader_class.return_value,
            ]
        )
        mocked_environment_class.assert_called_with(
            loader=mocked_choice_loader_class.return_value
        )
        mocked_check_template.assert_called_once_with(
            text_generator, "text", "text.ass"
        )

    def test_convert_icon(self):
        """Test the convertion of an available icon name to its code."""
        # create object
        text_generator = TextGenerator("package")
        text_generator.icon_map = {"music": "0xf001"}

        self.assertEqual(text_generator.convert_icon("music"), "\uf001")
        self.assertEqual(text_generator.convert_icon("other"), " ")

    def test_convert_icon_unavailable(self):
        """Test the convertion of an unavailable icon name to a generic code."""
        # create object
        text_generator = TextGenerator("package")

        self.assertEqual(text_generator.convert_icon("unavailable"), " ")

    def test_convert_icon_none(self):
        """Test the convertion of a null icon name is handled."""
        # create object
        text_generator = TextGenerator("package")

        self.assertEqual(text_generator.convert_icon(None), "")

    def test_convert_link_type_name(self):
        """Test the convertion of a link type to its long name."""
        self.assertEqual(TextGenerator.convert_link_type_name("OP"), "Opening")
        self.assertEqual(TextGenerator.convert_link_type_name("ED"), "Ending")
        self.assertEqual(TextGenerator.convert_link_type_name("IN"), "Insert song")
        self.assertEqual(TextGenerator.convert_link_type_name("IS"), "Image song")

    def test_convert_duration(self):
        """Test the convertion of durations."""
        self.assertEqual(TextGenerator.convert_duration(42), "0:42")
        self.assertEqual(TextGenerator.convert_duration(425), "7:05")
        self.assertEqual(TextGenerator.convert_duration(738), "12:18")
        self.assertEqual(TextGenerator.convert_duration(3618), "1:00:18")

    def test_convert_duration_undefined(self):
        """Test the convertion of undefined durations."""
        self.assertEqual(TextGenerator.convert_duration(None), "")
        self.assertEqual(TextGenerator.convert_duration("undefined"), "undefined")

    @patch.object(TextGenerator, "get_environment_loaders", autospec=True)
    def test_check_template_custom(self, mocked_get_environment_loaders):
        """Test to find a custom template."""
        mocked_loader_custom = MagicMock()
        mocked_loader_custom.list_templates.return_value = ["idle.ass"]
        mocked_loader_default = MagicMock()
        mocked_loader_default.list_templates.return_value = ["idle.ass"]

        mocked_get_environment_loaders.return_value = [
            mocked_loader_custom,
            mocked_loader_default,
        ]

        with self.assertLogs("dakara_player.text", "DEBUG") as logger:
            text_generator = TextGenerator("package", filenames={"idle": "idle.ass"})
            text_generator.check_template("idle", "idle.ass")

        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.text:Loading custom idle text "
                "template file 'idle.ass'"
            ],
        )

    @patch.object(TextGenerator, "get_environment_loaders", autospec=True)
    def test_check_template_default(self, mocked_get_environment_loaders):
        """Test to find a default template."""
        mocked_loader_custom = MagicMock()
        mocked_loader_custom.list_templates.return_value = []
        mocked_loader_default = MagicMock()
        mocked_loader_default.list_templates.return_value = ["idle.ass"]

        mocked_get_environment_loaders.return_value = [
            mocked_loader_custom,
            mocked_loader_default,
        ]

        with self.assertLogs("dakara_player.text", "DEBUG") as logger:
            text_generator = TextGenerator("package", filenames={"idle": "idle.ass"})
            text_generator.check_template("idle", "idle.ass")

        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.text:Loading default idle "
                "text template file 'idle.ass'"
            ],
        )

    @patch.object(TextGenerator, "get_environment_loaders", autospec=True)
    def test_check_template_not_found(self, mocked_get_environment_loaders):
        """Test to find an unaccessible template."""
        mocked_loader_custom = MagicMock()
        mocked_loader_custom.list_templates.return_value = []
        mocked_loader_default = MagicMock()
        mocked_loader_default.list_templates.return_value = []

        mocked_get_environment_loaders.return_value = [
            mocked_loader_custom,
            mocked_loader_default,
        ]

        with self.assertRaisesRegex(
            TemplateNotFoundError, "No idle text template file found for 'idle.ass'"
        ):
            text_generator = TextGenerator("package", filenames={"idle": "idle.ass"})
            text_generator.check_template("idle", "idle.ass")


class TextGeneratorIntegrationTestCase(TestCase):
    """Test the text generator class in real conditions."""

    def setUp(self):
        # create info dictionary
        self.idle_info = {"notes": ["VLC 0.0.0", "Dakara player 0.0.0"]}

        # create playlist entry
        self.playlist_entry = {
            "song": {
                "title": "Song title",
                "duration": 263,
                "artists": [{"name": "Artist name"}, {"name": "Other artist"}],
                "works": [
                    {
                        "work": {
                            "title": "Work title",
                            "subtitle": "Subtitle of the work",
                            "work_type": {
                                "name": "Work type name",
                                "icon_name": "music",
                            },
                        },
                        "link_type": "OP",
                        "link_type_number": 1,
                        "episodes": "1, 2, 3",
                    }
                ],
                "file_path": "path/of/the/file",
            },
            "owner": {"username": "User"},
            "use_instrumental": True,
            "date_created": "1970-01-01T00:00:00.00",
        }

        # create text generator object
        self.text_generator = TextGenerator(
            package="dakara_player.resources.templates",
            filenames={"idle": "idle.ass", "transition": "transition.ass"},
        )
        self.text_generator.load()

    def test_load_templates_default(self):
        """Test to load default templates using an existing directory.

        Integration test.
        """
        with TemporaryDirectory() as temp_str:
            temp = Path(temp_str)

            # create object
            text_generator = TextGenerator(
                package="dakara_player.resources.templates",
                directory=temp,
                filenames={"idle": "idle.ass", "transition": "transition.ass"},
            )

            # call the method
            text_generator.load_templates()

            # assert there are templates defined
            loader_custom, loader_default = text_generator.environment.loader.loaders
            self.assertNotIn("idle.ass", loader_custom.list_templates())
            self.assertNotIn("transition.ass", loader_custom.list_templates())
            self.assertIn("idle.ass", loader_default.list_templates())
            self.assertIn("transition.ass", loader_default.list_templates())

    def test_load_templates_custom(self):
        """Test to load custom templates using an existing directory."""
        with TemporaryDirectory() as temp_str:
            temp = Path(temp_str)

            # prepare directory
            with as_file(
                files("dakara_player.resources.templates").joinpath("idle.ass")
            ) as file:
                copy(file, temp)

            with as_file(
                files("dakara_player.resources.templates").joinpath("transition.ass")
            ) as file:
                copy(file, temp)

            # create object
            text_generator = TextGenerator(
                package="dakara_player.resources.templates",
                directory=temp,
                filenames={"idle": "idle.ass", "transition": "transition.ass"},
            )

            # call the method
            text_generator.load_templates()

            # assert there are templates defined
            loader_custom, loader_default = text_generator.environment.loader.loaders
            self.assertIn("idle.ass", loader_custom.list_templates())
            self.assertIn("transition.ass", loader_custom.list_templates())

    def test_get_idle_text(self):
        """Test the generation of an idle text."""
        # call method
        result = self.text_generator.get_text("idle", self.idle_info)

        # check file content
        with as_file(files("tests.resources").joinpath("idle.ass")) as file:
            idle_text_content = file.read_text(encoding="utf8")
            self.assertEqual(idle_text_content.rstrip(), result.rstrip())

    def test_get_transition_text(self):
        """Test the generation of a transition text."""
        # call method
        result = self.text_generator.get_text(
            "transition", {"playlist_entry": self.playlist_entry, "fade_in": True}
        )

        # check file content
        with as_file(files("tests.resources").joinpath("transition.ass")) as file:
            transition_text_content = file.read_text(encoding="utf8")
            self.assertEqual(transition_text_content.rstrip(), result.rstrip())


class SeparatePackageLastDirectoryTestCase(TestCase):
    """Test the separate_package_last_directory function."""

    def test_three(self):
        """Test separate package with three elements."""
        package, directory = separate_package_last_directory("a.b.c")
        self.assertEqual(package, "a.b")
        self.assertEqual(directory, "c")

    def test_two(self):
        """Test separate package with two elements."""
        package, directory = separate_package_last_directory("a.b")
        self.assertEqual(package, "a")
        self.assertEqual(directory, "b")

    def test_one(self):
        """Test separate package with one element."""
        package, directory = separate_package_last_directory("a")
        self.assertEqual(package, "")
        self.assertEqual(directory, "a")



"""
Dakara player.

Interface between the Dakara server and a media player, for the Dakara project.
"""

from dakara_player import (
    audio,
    background,
    font,
    manager,
    media_player,
    mrl,
    player,
    text,
    user_resources,
    version,
    web_client,
    window,
)
from dakara_player.version import __date__, __version__

__all__ = [
    "audio",
    "manager",
    "mrl",
    "text",
    "version",
    "window",
    "background",
    "font",
    "player",
    "user_resources",
    "web_client",
    "media_player",
    "__version__",
    "__date__",
]



"""Player."""

import logging
from contextlib import ExitStack
from pathlib import Path
from tempfile import TemporaryDirectory

from dakara_base.exceptions import DakaraError
from dakara_base.safe_workers import Runner, WorkerSafeThread

from dakara_player.font import get_font_loader_class
from dakara_player.manager import DakaraManager
from dakara_player.media_player.mpv import MediaPlayerMpv
from dakara_player.media_player.vlc import MediaPlayerVlc
from dakara_player.version import check_version
from dakara_player.web_client import HTTPClientDakara, WebSocketClientDakara

FontLoader = get_font_loader_class()

MEDIA_PLAYER_CLASSES = {
    "mpv": MediaPlayerMpv.from_version,
    "vlc": MediaPlayerVlc,
}

logger = logging.getLogger(__name__)


class DakaraPlayer(Runner):
    """Class associated with the main thread.

    It simply starts, launchs the worker and waits for it to terminate or for a
    user Ctrl+C to be fired.
    """

    def init_runner(self, config):
        """Initialization.

        Creates the worker stop event.

        Args:
            config (dict): Configuration for the program.
        """
        # store arguments
        self.config = config

        # inform the user
        logger.debug("Started main")

    def load(self):
        """Execute side-effect actions."""
        # check version
        check_version()

    def run(self):
        """Launch the worker and wait for the end."""
        self.run_safe(DakaraWorker, self.config)


class DakaraWorker(WorkerSafeThread):
    """Class associated with the worker thread.

    It simply starts, loads configuration, set the different worker, launches
    the main thread and waits for the end.
    """

    def init_worker(self, config):
        """Initialization.

        Load the config and set the logger loglevel.

        Args:
            config (dict): Configuration for the program.
        """
        self.config = config

        # set thread
        self.thread = self.create_thread(target=self.run)

        # inform the user
        logger.debug("Starting Dakara worker")

    def get_media_player_class(self):
        """Get the class of the requested media player.

        Fallback to VLC if none was provided in config. If the requested media
        player is not known, raise an error.

        Returns:
            dakara_player.media_player.base.MediaPlayer: Specialized class of
            the media player.
        """
        media_player_name = self.config["player"].get("player_name", "vlc")

        try:
            return MEDIA_PLAYER_CLASSES[media_player_name.lower()]

        except KeyError as error:
            raise UnsupportedMediaPlayerError(
                "No media player for '{}'".format(media_player_name)
            ) from error

    def run(self):
        """Worker main method.

        It sets up the different workers and uses them as context managers,
        which guarantee that their different clean methods will be called
        prorperly.

        Then it starts the polling thread and waits for the end.

        When `run` is called, the end can come for several reasons:

        * the main thread (who calls the worker thread) has caught a Ctrl+C
          from the user;
        * an exception has been raised within the `run` method (directly in
          the worker thread);
        * an exception has been raised within the polling thread.
        """
        # get the different workers as context managers
        # ExitStack makes the management of multiple context managers simpler
        # This mechanism plus the use of Worker classes allow to gracelly end
        # the execution of any thread within the context manager. It guarantees
        # as well that on leaving this context manager, all cleanup tasks will
        # be executed.
        with ExitStack() as stack:
            # temporary directory
            tempdir = Path(stack.enter_context(TemporaryDirectory(suffix=".dakara")))

            # font loader
            font_loader = stack.enter_context(
                FontLoader("dakara_player.resources.fonts")
            )
            font_loader.load()

            # media player
            media_player = stack.enter_context(
                self.get_media_player_class()(
                    self.stop,
                    self.errors,
                    self.config["player"],
                    tempdir,
                )
            )
            media_player.load()

            # communication with the dakara HTTP server
            http_client = HTTPClientDakara(
                self.config["server"], endpoint_prefix="api/", mute_raise=True
            )
            http_client.load()
            http_client.authenticate()
            token_header = http_client.get_token_header()

            # communication with the dakara WebSocket server
            websocket_client = stack.enter_context(
                WebSocketClientDakara(
                    self.stop,
                    self.errors,
                    self.config["server"],
                    header=token_header,
                    endpoint="ws/playlist/device/",
                )
            )

            # manager for the precedent workers
            dakara_manager = DakaraManager(  # noqa F841
                font_loader, media_player, http_client, websocket_client
            )

            # start the worker timer
            websocket_client.timer.start()

            # wait for stop event
            self.stop.wait()

            # leaving this method means leaving all the context managers and
            # stopping the program


class UnsupportedMediaPlayerError(DakaraError):
    """Raised if an unknown media player is requested."""



from unittest import TestCase
from unittest.mock import MagicMock

from dakara_player.manager import DakaraManager


class DakaraManagerTestCase(TestCase):
    """Test the dakara manager class."""

    def setUp(self):
        # create a mock font loader
        self.font_loader = MagicMock()

        # create a mock VLC player
        self.media_player = MagicMock()

        # create a mock Dakara HTTP server
        self.client_http = MagicMock()

        # create a mock Dakara websocket server
        self.client_websocket = MagicMock()

        # create a Dakara manager
        self.dakara_manager = DakaraManager(
            self.font_loader,
            self.media_player,
            self.client_http,
            self.client_websocket,
        )

    def test_start_idle(self):
        """Test to launch the dakara manager when there is nothing to play."""
        # call the methods and prevent to run as thread
        self.dakara_manager.play_idle_screen()

        # call assertions
        self.media_player.play.assert_called_once_with("idle")
        self.media_player.set_playlist_entry.assert_not_called()

    def test_start_play_playlist_entry(self):
        """Test to launch the dakara manager when there is something to play."""
        playlist_entry = {"id": 42}

        # call the methods and prevent to run as thread
        self.dakara_manager.play_playlist_entry(playlist_entry)

        # call assertions
        self.media_player.play.assert_not_called()
        self.media_player.set_playlist_entry.assert_called_once_with(playlist_entry)

    def test_handle_error(self):
        """Test the callback called on error."""
        # call the method
        self.dakara_manager.handle_error(999, "message")

        # call assertions
        self.client_http.post_player_error.assert_called_once_with(999, "message")

    def test_handle_finished(self):
        """Test the callback called on song end."""
        # call the method
        self.dakara_manager.handle_finished(999)

        # call assertions
        self.client_http.put_status_finished.assert_called_once_with(999)

    def test_handle_started_transition(self):
        """Test the callback called on transition start."""
        # call the method
        self.dakara_manager.handle_started_transition(999)

        # call assertions
        self.client_http.put_status_started_transition.assert_called_once_with(999)

    def test_handle_started_song(self):
        """Test the callback called on song start."""
        # call the method
        self.dakara_manager.handle_started_song(999)

        # call assertions
        self.client_http.put_status_started_song.assert_called_once_with(999)

    def test_handle_could_not_play(self):
        """Test the callback called when a playlist entry could not play."""
        # call the method
        self.dakara_manager.handle_could_not_play(999)

        # call assertions
        self.client_http.put_status_could_not_play.assert_called_once_with(999)

    def test_handle_paused(self):
        """Test the callback called when the player is paused."""
        # call the method
        self.dakara_manager.handle_paused(999, 10)

        # call assertions
        self.client_http.put_status_paused.assert_called_once_with(999, 10)

    def test_handle_resumed(self):
        """Test the callback called when the player resumed playing."""
        # call the method
        self.dakara_manager.handle_resumed(999, 10)

        # call assertions
        self.client_http.put_status_resumed.assert_called_once_with(999, 10)

    def test_handle_updated_timing(self):
        """Test the callback called when the player updated its timing."""
        # call the method
        self.dakara_manager.handle_updated_timing(999, 10)

        # call assertions
        self.client_http.put_status_update_timing.assert_called_once_with(999, 10)

    def test_do_command_successful(self):
        """Test the command manager for valid commands."""
        # mock the playlist entry ID
        self.dakara_manager.media_player.playing_id = 42

        # call the method for pause
        self.dakara_manager.do_command("pause")

        # assert the call
        self.media_player.pause.assert_called_with()

        # reset the mock
        self.media_player.pause.reset_mock()

        # call the method for resume
        self.dakara_manager.do_command("resume")

        # assert the call
        self.media_player.resume.assert_called_with()

        # call the method for skip
        self.dakara_manager.do_command("skip")

        # assert the call
        self.media_player.skip.assert_called_with()

    def test_do_command_error(self):
        """Test the command manager for an invalid command."""
        # call the method
        with self.assertRaises(AssertionError):
            self.dakara_manager.do_command("invalid")



"""Base objects for media players."""

import logging
from abc import ABC, abstractmethod
from functools import wraps
from pathlib import Path
from threading import Timer

from dakara_base.directory import directories
from dakara_base.exceptions import DakaraError
from dakara_base.safe_workers import Worker

from dakara_player.audio import get_audio_files
from dakara_player.background import BackgroundLoader
from dakara_player.text import TextGenerator
from dakara_player.version import __version__

TRANSITION_BG_NAME = "transition.png"
TRANSITION_TEXT_NAME = "transition.ass"
TRANSITION_DURATION = 2

IDLE_BG_NAME = "idle.png"
IDLE_TEXT_NAME = "idle.ass"
IDLE_DURATION = 300

PLAYER_CLOSING_DURATION = 3

REWIND_FAST_FORWARD_DURATION = 10


logger = logging.getLogger(__name__)


class MediaPlayer(Worker, ABC):
    """Abstract class to manipulate a media player.

    The class can be used as a context manager that closes the media player
    automatically on exit.

    Args:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.
        config (dict): Dictionary of configuration.
        tempdir (pathlib.Path): Path of the temporary directory.

    Attributes:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.
        player_name (str): Name of the media player.
        fullscreen (bool): If `True`, the media player will be fullscreen.
        kara_folder_path (pathlib.Path): Path to the karaoke folder.
        playlist_entry (dict): Playlist entyr object.
        callbacks (dict): High level callbacks associated with the media
            player.
        warn_long_exit (bool): If `True`, display a warning message if the media
            player takes too long to stop.
        durations (dict of int): Duration of the different screens in seconds.
        text_paths (dict of pathlib.Path): Path of the different text screens.
        text_generator (dakara_player.text.TextGenerator): Text
            generator instance.
        background_loader
        (dakara_player.background.BackgroundLoader): Background
            loader instance.
    """

    player_name = None

    @staticmethod
    @abstractmethod
    def is_available():
        """Indicate if the implementation is available.

        Must be overriden.

        Returns:
            bool: `True` if the media player is useable.
        """

    def init_worker(self, config, tempdir, warn_long_exit=True):
        """Initialize the base objects of the media player.

        Actions performed in this method should not have any side effects
        (query file system, etc.).

        Args:
            config (dict): Dictionary of configuration.
            tempdir (pathlib.Path): Path of the temporary directory.
            warn_long_exit (bool): If `True`, the class will display a warning
                message if the media player takes too long to stop.
        """
        self.check_is_available()

        # karaoke parameters
        self.fullscreen = config.get("fullscreen", False)
        self.kara_folder_path = Path(config.get("kara_folder", ""))

        # inner objects
        self.playlist_entry = None
        self.callbacks = {}
        self.warn_long_exit = warn_long_exit

        # set durations
        config_durations = config.get("durations") or {}
        self.durations = {
            "idle": IDLE_DURATION,
            "transition": config_durations.get(
                "transition_duration", TRANSITION_DURATION
            ),
            "rewind_fast_forward": config_durations.get(
                "rewind_fast_forward_duration", REWIND_FAST_FORWARD_DURATION
            ),
        }

        # set text paths
        self.text_paths = {
            "idle": tempdir / IDLE_TEXT_NAME,
            "transition": tempdir / TRANSITION_TEXT_NAME,
        }

        # set text generator
        config_texts = config.get("templates") or {}
        self.text_generator = TextGenerator(
            package="dakara_player.resources.templates",
            directory=directories.user_data_path / "player" / "templates",
            filenames={
                "transition": config_texts.get(
                    "transition_template_name", TRANSITION_TEXT_NAME
                ),
                "idle": config_texts.get("idle_template_name", IDLE_TEXT_NAME),
            },
        )

        # set background loader
        config_backgrounds = config.get("backgrounds") or {}
        self.background_loader = BackgroundLoader(
            destination=tempdir,
            package="dakara_player.resources.backgrounds",
            directory=directories.user_data_path / "player" / "backgrounds",
            filenames={
                "transition": config_backgrounds.get(
                    "transition_background_name", TRANSITION_BG_NAME
                ),
                "idle": config_backgrounds.get("idle_background_name", IDLE_BG_NAME),
            },
        )

        # set default callbacks
        self.set_default_callbacks()

        # call specialized constructor
        self.init_player(config, tempdir)

    def init_player(self, config, tempdir):
        """Initialize the objects of the specific media player.

        Actions performed in this method should not have any side effects
        (query file system, etc.).

        Can be overriden.

        Args:
            config (dict): Dictionary of configuration.
            tempdir (pathlib.Path): Path of the temporary directory.
        """

    def load(self):
        """Perform base actions with side effects for media player initialization."""
        # check kara folder
        self.check_kara_folder_path()

        # load text generator
        self.text_generator.load()

        # load backgrounds
        self.background_loader.load()

        self.load_player()

    def load_player(self):
        """Perform actions with side effects for specialized media player
        initialization.

        Can be overriden.
        """

    @abstractmethod
    def get_timing(self):
        """Get media player timing.

        Must be overriden.

        Returns:
            int: Current song timing in seconds if a song is playing, or 0 when
                idle or during transition screen.
        """

    @staticmethod
    @abstractmethod
    def get_version():
        """Get media player version.

        Must be overriden.

        Returns:
            packaging.version.Version: Parsed version of the media player.
        """

    @classmethod
    def get_version_str(cls):
        """Get media player formatted version.

        Returns:
            str: Formatted version of the media player. If the version
            cannot be obtained, display `unknown version`.
        """
        try:
            return str(cls.get_version())

        except VersionNotFoundError:
            return "unknown version"

    @abstractmethod
    def is_playing(self):
        """Query if the media player is playing something.

        Must be overriden.

        Returns:
            bool: `True` if the media player is playing something.
        """

    @abstractmethod
    def is_paused(self):
        """Query if the media player is paused.

        Must be overriden.

        Returns:
            bool: `True` if the media player is paused.
        """

    @abstractmethod
    def is_playing_this(self, what):
        """Query if the media player is playing the requested media type.

        Must be overriden.

        Args:
            what (str): Tell if the media player current track is of the
                requested type, but not if it is actually playing it (it can be
                in pause).

        Returns:
            bool: `True` if the media player is playing the requested type.
        """

    @abstractmethod
    def play(self, what):
        """Request the media player to play something.

        No preparation should be done by this function, i.e. the media track
        should have been prepared already by `set_playlist_entry`.

        Must be overriden.

        Args:
            what (str): What media to play.
        """

    @abstractmethod
    def pause(self):
        """Request the media player to pause.

        Can only work on transition screens or songs. Pausing should have no
        effect if the media player is already paused.

        Must be overriden.
        """

    @abstractmethod
    def resume(self):
        """Request the media player to resume playing.

        Can only work on transition screens or songs. Resuming should have no
        effect if the media player is already playing.

        Must be overriden.
        """

    @abstractmethod
    def restart(self):
        """Request to restart the current media.

        Can only work on songs.

        Must be overriden.
        """

    @abstractmethod
    def skip(self, no_callback=False):
        """Request to skip the current media.

        Can only work on transition screens or songs. The media player should
        continue playing, but media has to be considered already finished.

        Args:
            no_callback (bool): If `True`, no callback to signal the song has
                finished will be executed.

        Must be overriden.
        """

    @abstractmethod
    def rewind(self):
        """Request to rewind a few seconds the media.

        Can only work on songs. It cannot rewind before the beginning of the media.

        Must be overriden.
        """

    def fast_forward(self):
        """Request to fast forward a few seconds the media.

        Can only work on songs. It cannot advance passed the end of the media.

        Must be overriden.
        """

    @abstractmethod
    def stop_player():
        """Request to stop the media player.

        Must be overriden.
        """

    def set_playlist_entry(self, playlist_entry, autoplay=True):
        """Prepare playlist entry base data to be played.

        Check if the song file exists, otherwise consider the song cannot be
        played.

        Args:
            playlist_entry (dict): Playlist entry object.
            autoplay (bool): If `True`, start to play transition screen as soon
                as possible.
        """
        file_path = self.kara_folder_path / playlist_entry["song"]["file_path"]

        if not file_path.is_file():
            logger.error("File not found '%s'", file_path)
            self.callbacks["error"](playlist_entry["id"], "File not found")
            self.callbacks["could_not_play"](playlist_entry["id"])
            return

        self.playlist_entry = playlist_entry

        self.set_playlist_entry_player(playlist_entry, file_path, autoplay)

    @abstractmethod
    def set_playlist_entry_player(self, playlist_entry, file_path, autoplay):
        """Prepare playlist entry data to be played.

        Prepare all media objects, subtitles, etc. for being played, for the
        transition screen and the song. Such data should be stored on a
        dedicated object, like `playlist_entry_data`.

        Must be overriden.

        Args:
            playlist_entry (dict): Playlist entry object.
            file_path (pathlib.Path): Absolute path to the song file.
            autoplay (bool): If `True`, start to play transition screen as soon
                as possible (i.e. as soon as the transition screen media is
                ready). The song media is prepared when the transition screen
                is playing.
        """

    def clear_playlist_entry(self):
        """Clean playlist entry base data."""
        logger.debug("Clearing internal memory")
        self.playlist_entry = None

        self.clear_playlist_entry_player()

    @abstractmethod
    def clear_playlist_entry_player(self):
        """Clean playlist entry data after being played.

        Must be overriden.
        """

    def set_callback(self, name, callback):
        """Set callback to the media player.

        Args:
            name (str): Name of the callback.
            callback (function): Callback.
        """
        self.callbacks[name] = callback

    @staticmethod
    def get_instrumental_file(filepath):
        """Get the instrumental audio file associated to a given song file.

        Consider that this instrumental file should be the only one audio file found.

        Args:
            filepath (pathlib.Path): Path to the media file.

        Returns:
            pathlib.Path: Path to the instrumental file. None if not found.
        """
        audio_files = get_audio_files(filepath)

        # accept only one audio file
        if len(audio_files) == 1:
            return audio_files[0]

        # otherwise return None
        return None

    def check_kara_folder_path(self):
        """Check if the karaoke folder exists."""
        if not self.kara_folder_path.exists():
            raise KaraFolderNotFound(
                'Karaoke folder "{}" does not exist'.format(self.kara_folder_path)
            )

    def check_is_available(self):
        """Check if the media player is installed and useable."""
        # check the target player is available
        if not self.is_available():
            raise MediaPlayerNotAvailableError(
                "{} is not available".format(self.player_name)
            )

    def set_default_callbacks(self):
        """Set dummy callbacks that have to be defined externally."""
        self.set_callback("started_transition", lambda playlist_entry_id: None)
        self.set_callback("started_song", lambda playlist_entry_id: None)
        self.set_callback("could_not_play", lambda playlist_entry_id: None)
        self.set_callback("finished", lambda playlist_entry_id: None)
        self.set_callback("paused", lambda playlist_entry_id, timing: None)
        self.set_callback("resumed", lambda playlist_entry_id, timing: None)
        self.set_callback("error", lambda playlist_entry_id, message: None)
        self.set_callback("updated_timing", lambda playlist_entry_id, timing: None)

    def exit_worker(self, *args, **kwargs):
        """Exit the worker.

        If `warn_long_exit` was `True` during initialization, send a warning
        after `PLAYER_CLOSING_DURATION` seconds if the worker is not closed
        yet.
        """
        if self.warn_long_exit:
            # send a warning within if the player has not stopped already
            timer_stop_player_too_long = Timer(
                PLAYER_CLOSING_DURATION, self.warn_stop_player_too_long
            )
            timer_stop_player_too_long.start()

        # stop player
        self.stop_player()

        if self.warn_long_exit:
            # clear the warning
            timer_stop_player_too_long.cancel()

    @classmethod
    def warn_stop_player_too_long(cls):
        """Notify the user that the player takes too long to stop."""
        logger.warning("{} takes too long to stop".format(cls.player_name))

    def generate_text(self, what, **kwargs):
        """Generate text screens for the requested action.

        Args:
            what (str): What text screen to generate.
            kwargs (dict): Extra arguments passed to the different text
                screens. For transition screen, accepts the `fade_in` boolean,
                which is `True` by default. If `False`, the transition screen will
                have no fade in effect.

        Returns:
            pathlib.Path: Path of the text screen.

        Raises:
            ValueError: If the type of screen to generate is unknown.
        """
        if what == "idle":
            text = self.text_generator.get_text(
                "idle",
                {
                    "notes": [
                        "{} {}".format(self.player_name, self.get_version_str()),
                        "Dakara player {}".format(__version__),
                    ]
                },
            )

        elif what == "transition":
            text = self.text_generator.get_text(
                "transition",
                {
                    "playlist_entry": self.playlist_entry,
                    "fade_in": kwargs.get("fade_in", True),
                },
            )

        else:
            raise ValueError("Unexpected action to generate text for: {}".format(what))

        self.text_paths[what].write_text(text, encoding="utf-8")

        return self.text_paths[what]


def on_playing_this(what_list, default_return=None):
    """Decorator for methods that necessitate the player to be playing
    something specifically.

    Args:
        what_list (list): List of possible states.
        default_return (any): Value to return if the state is different.
            Default to `None`.
    """

    def decorator(function):
        @wraps(function)
        def wrap(self, *args, **kwargs):
            for what in what_list:
                if self.is_playing_this(what):
                    break

            else:
                return default_return

            return function(self, *args, **kwargs)

        return wrap

    return decorator


class KaraFolderNotFound(DakaraError):
    """Error raised when the kara folder cannot be found."""


class MediaPlayerNotAvailableError(DakaraError):
    """Error raised when trying to use a target player that cannot be found."""


class InvalidStateError(RuntimeError):
    """Error raised when the state of the player is invalid."""


class VersionNotFoundError(Exception):
    """Error raised when the version of the player cannot be found."""



import sys
from pathlib import Path
from unittest import TestCase, skipIf

from dakara_player.mrl import mrl_to_path, path_to_mrl


class MrlFunctionsTestCase(TestCase):
    """Test the MRL conversion functions."""

    is_windows = sys.platform.startswith("win")

    @skipIf(is_windows, "Tested on POSIX")
    def test_mrl_to_path_posix(self):
        """Test to convert MRL to path for POSIX."""
        path = mrl_to_path("file:///home/username/directory/file%20name.ext")
        self.assertEqual(
            path, Path("/home").resolve() / "username" / "directory" / "file name.ext"
        )

    @skipIf(not is_windows, "Tested on Windows")
    def test_mrl_to_path_windows(self):
        """Test to convert MRL to path for Windows."""
        path = mrl_to_path("file:///C:/Users/username/directory/file%20name.ext")
        self.assertEqual(
            path,
            Path("C:/Users").resolve() / "username" / "directory" / "file name.ext",
        )

    @skipIf(is_windows, "Tested on POSIX")
    def test_path_to_mrl_posix(self):
        """Test to convert path to MRL for POSIX."""
        mrl = path_to_mrl(
            Path("/home").resolve() / "username" / "directory" / "file name.ext"
        )
        self.assertEqual(mrl, "file:///home/username/directory/file%20name.ext")

    @skipIf(not is_windows, "Tested on Windows")
    def test_path_to_mrl_windows(self):
        """Test to convert path to MRL for Windows."""
        mrl = path_to_mrl(
            Path("C:/Users").resolve() / "username" / "directory" / "file name.ext"
        )
        self.assertEqual(mrl, "file:///C:/Users/username/directory/file%20name.ext")
